#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

# Notice: Most recent documentation is available at doc/test_class.py.template.

import os, sys, re, urlparse, urlgrabber
import autoqa.util
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autotest_lib.client.bin import utils
from autoqa.repoinfo import repoinfo
from autoqa.bodhi_utils import bodhi_list, bodhi_post_testresult, bodhi_already_commented, query_update
from autoqa.koji_utils import SimpleKojiClientSession

### FIXME: AutoQA should provide testarch
### https://fedorahosted.org/autoqa/ticket/272
localarch = autoqa.util.get_basearch()
testarch = localarch

# TODO: move these into koji_utils
def list_pending_builds(repotag):
    '''List pending builds for the repo associated with the given koji tag'''
    pending_tag = repotag + '-pending'
    koji = SimpleKojiClientSession()
    builds = koji.listTagged(pending_tag)
    nvrs = sorted([build['nvr'] for build in builds])
    return nvrs

def strip_epoch(rpmstr):
    return rpmstr.split(':',1)[-1] # works for E:N-V-R.A or N-V-R.A
def strip_arch(rpmstr):
    return rpmstr.rsplit('.',1)[0]

# --- Everything below here is depcheck-specific

class depcheck(AutoQATest):
    version = 2 # increment this if setup() changes

    @ExceptionCatcher()
    def setup(self, *args, **kwargs):
        utils.system('yum -y install python-rpmfluff mash')

    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.result = 'PASSED'
        self.highlights = []
        # dict of problematic NVRs - can't find in Bodhi
        # nvr -> error
        self.problematic = dict()

    def is_accepted(self, update, arch=testarch):
        '''Return True if the given update has been marked as Accepted by a
        previous depcheck run, False otherwise.'''
        try:
            (result, time) = bodhi_already_commented(update, 'depcheck', arch)
            return result == 'PASSED'
        except ValueError as e:
            msg = 'ERROR: %s' % e
            print msg
            self.highlights.append(msg)
            self.problematic[update] = e
            return False

    @ExceptionCatcher()
    def run_once(self, nvrs, kojitag, **kwargs):
        super(self.__class__, self).run_once()

        # Get our inputs
        pending = set(list_pending_builds(kojitag))
        # XXX set testarch to noarch for noarch updates?
        accepted_nvrs = set([nvr for nvr in pending if self.is_accepted(nvr)])
        print 'Accepted builds: %s' % sorted(accepted_nvrs)
        pending_nvrs = sorted(pending.difference(accepted_nvrs).difference(self.problematic.keys()))
        print 'Pending builds: %s' % pending_nvrs
        print 'Problematic builds: %s' % self.problematic.keys()

        # Fetch packages and build commandline
        repo = repoinfo.getrepo_by_tag(kojitag)
        repoid = repo['name']
        cmd = './depcheck %s' % repoid
        for (nvr, rpms) in self.fetch_nvrs(accepted_nvrs, self.tmpdir):
            for rpm in rpms:
                cmd += ' -a %s' % rpm
        pending_parent_map = {}
        for (nvr, rpms) in self.fetch_nvrs(pending_nvrs, self.tmpdir):
            for rpm in rpms:
                cmd += ' %s' % rpm
                # While we're here, construct subpackage NVR -> parent NVR map
                # package filenames are all: /some/path/to/N-V-R.A.rpm
                this_nvr = os.path.basename(strip_arch(rpm[:-4]))
                pending_parent_map[this_nvr] = nvr

        # Run the test
        self.outputs = utils.system_output(cmd, retain_output=True)

        # Process output
        results = dict(IGNORE=[], REJECT=[], ACCEPT=[])
        data_re = re.compile('^(IGNORE|REJECT|ACCEPT): (.*)$')
        for line in self.outputs.splitlines():
            match = data_re.match(line)
            if not match:
                continue
            (what, builds) = match.groups()
            results[what] = builds.split()

        # We need to split ENVRS to the groups according to update titles
        # since one update may contain more envrs, and we want to report
        # results for updates, not envrs
        def envra_to_nvr(envra):
            return strip_arch(strip_epoch(envra))


        accepted_builds = set()
        for envra in results['ACCEPT']:
            nvr = envra_to_nvr(envra)
            if pending_parent_map.has_key(nvr):
                accepted_builds.add(pending_parent_map[nvr])
            else:
                print "Build accepted but not in pending updates, skipping: %s" % envra

        rejected_builds = set()
        for envra in results['REJECT']:
            nvr = envra_to_nvr(envra)
            if pending_parent_map.has_key(nvr):
                rejected_builds.add(pending_parent_map[nvr])
            else:
                print "Build rejected but not in pending updates, skipping: %s" % envra

        updates = {}
        all_builds = []
        all_builds.extend(accepted_builds)
        all_builds.extend(rejected_builds)
        for nvr in all_builds:
            bodhi_info = None
            bodhi_info = query_update(nvr)
            if bodhi_info is None:
                print >> sys.stderr, "Update containing %s not found in Bodhi" % nvr
                continue

            if bodhi_info['title'] not in updates.keys():
                updates[bodhi_info['title']] = bodhi_info

        # For every update in tested updates, test if every NVR is OK. If
        # at least one NVR from the update is not OK, then update is rejected
        for update in updates.values():
            result = 'PASSED'
            for build in update['builds']:
                if build['nvr'] not in accepted_builds:
                    result = 'FAILED'
                    self.result = 'FAILED'
                    break
            print "Posting to bodhi: %s %s" % (update['title'], result)
            bodhi_post_testresult(update['title'], 'depcheck', result, self.autotest_url, testarch, karma=0)

        if self.result == 'PASSED' and len(self.problematic) > 0:
            print >> sys.stderr, 'Problematic builds found, setting status to NEEDS_INSPECTION'
            self.result = 'NEEDS_INSPECTION'

        # TODO: force filling up reports for updates (resultsdb stuff)
        self.highlights.append('ACCEPTED:')
        self.highlights.extend(['  ' + build for build in results['ACCEPT']])
        self.highlights.append('REJECTED:')
        self.highlights.extend(['  ' + build for build in results['REJECT']])
        self.summary = kojitag

    def fetch_nvrs(self, nvrs, localdir):
        '''Fetch all the RPMs for the given package NVRs.
        Return a list of tuples: [(nvr,rpmlist), ...]'''
        # NOTE: Currently we're just fetching everything from koji. This might not
        # be the fastest thing, but it works for everyone, everywhere. If there's
        # a shortcut we can use on the autoqa test systems, we should use that,
        # but it needs to gracefully fall back to something like this.

        # TODO: Download only needed archs to speed it up
        # https://fedorahosted.org/pipermail/autoqa-devel/2011-February/001615.html
        # https://fedorahosted.org/autoqa/ticket/272
        rpms = []
        koji = SimpleKojiClientSession()
        for nvr in sorted(nvrs):
            pkgs = []
            # XXX deal with debuginfo/src RPMs?
            print "Fetching RPMs for %s" % nvr
            try:
                for rpm in koji.nvr_to_rpms(nvr, debuginfo=False, src=False):
                    urlobj = urlparse.urlparse(rpm['url'])
                    rpmname = os.path.basename(urlobj.path)
                    outpath = os.path.join(localdir, rpmname)
                    print "  %s" % rpmname
                    outfile = autoqa.util.grabber.urlgrab(rpm['url'], outpath)
                    pkgs.append(outfile)
            except urlgrabber.grabber.URLGrabError as e:
                self.result = 'ABORTED'
                raise
            print
            rpms.append((nvr, pkgs))
        return rpms
